var Events = require("./events"), urlUtil = require("./../shared/url_util"), cmd = require("./../shared/service_msg_cmd"), uuid = require("node-uuid"), path = require("path"), fs = require("fs");

var URL_PATH_PREFIX = "/tcs", CLIENT_STATIC_SCRIPT = "/hammerhead.js", CLIENT_SCRIPT_FILE_PATH = path.join(__dirname, "../../_compiled_/hammerhead_client/hammerhead.js"), TASK_SCRIPT_PATHNAME = "/task.js", SERVICE_MSGS_PATHNAME = "/tccservicemsg", SETTINGS_TEMPLATE_PATH = path.join(__dirname, "../../_compiled_/hammerhead_client/settings.jstmpl");

var ServiceChannel = module.exports = function(proxyPort, proxyHostname, cookieShelf, staticResources) {
    this.events = new Events();
    this.cookieShelf = cookieShelf;
    this.port = proxyPort;
    this.hostname = proxyHostname;
    this.staticResourceDescriptors = {};
    this.urlPathPrefix = URL_PATH_PREFIX + uuid.v4();
    this.serviceMsgUrl = this._getServiceUrl(SERVICE_MSGS_PATHNAME);
    this.taskScriptUrl = this._getServiceUrl(TASK_SCRIPT_PATHNAME);
    this.staticCssUrl = null;
    this.staticScriptUrl = null;
    this.hammerheadScriptUrl = null;
    if (staticResources.script) {
        this.staticScriptUrl = this.registerStaticResource(staticResources.script.path, {
            contentType: staticResources.script.contentType || "application/x-javascript",
            content: staticResources.script.content
        });
    }
    if (staticResources.css) {
        this.staticCssUrl = this.registerStaticResource(staticResources.css.path, {
            contentType: staticResources.css.contentType || "/uistyle.css",
            content: staticResources.css.content
        });
    }
    if (staticResources.others) {
        for (var i = 0; i < staticResources.others.length; i++) {
            this.registerStaticResource(staticResources.others[i].path, {
                contentType: staticResources.others[i].contentType,
                content: staticResources.others[i].content
            });
        }
    }
    this._registerHammerheadScript();
};

ServiceChannel.prototype._registerHammerheadScript = function() {
    var fileContent = fs.readFileSync(CLIENT_SCRIPT_FILE_PATH).toString();
    this.hammerheadScriptUrl = this.registerStaticResource(CLIENT_STATIC_SCRIPT, {
        contentType: "application/x-javascript",
        content: fileContent
    });
};

ServiceChannel.prototype._getServiceUrl = function(resourcePathname) {
    return urlUtil.formatUrl({
        protocol: "http:",
        hostname: this.hostname,
        port: this.port,
        pathname: this.urlPathPrefix + resourcePathname
    });
};

ServiceChannel.prototype.getSettingsScript = function(vars) {
    var template = fs.readFileSync(SETTINGS_TEMPLATE_PATH).toString();
    return template.replace(/"<@\s*(\S+)\s*@>"/g, function(subStr, arg) {
        return vars && typeof vars[arg] !== "undefined" ? vars[arg] : subStr;
    });
};

ServiceChannel.prototype._processServiceMsg = function(ctx) {
    var msgJson = "";
    ctx.reqBodyChunks.forEach(function(chunk) {
        msgJson += chunk.toString();
    });
    var msg = null;
    try {
        msg = JSON.parse(msgJson);
    } catch (err) {
        ctx.res.statusCode = 500;
        ctx.res.end();
        return;
    }
    var sendServiceResponse = function(data) {
        if (data !== null && typeof data !== "undefined") {
            ctx.res.end(JSON.stringify(data));
        } else {
            ctx.res.end();
        }
    };
    if (msg.cmd === cmd.SET_COOKIE) {
        var serviceChannel = this, reqInfo = urlUtil.parseProxyUrl(msg.url), originUrl = reqInfo ? urlUtil.formatUrl(reqInfo.originResourceInfo) : msg.url, jobInfo = {
            uid: msg.jobUid,
            ownerToken: msg.jobOwnerToken
        };
        this.cookieShelf.setCookieByClient(jobInfo, originUrl, msg.cookie, function() {
            serviceChannel.cookieShelf.getClientCookieString(jobInfo, originUrl, sendServiceResponse);
        });
    } else {
        this.events.for(msg.jobOwnerToken).emit("serviceMsg", msg, sendServiceResponse);
    }
};

ServiceChannel.prototype.shouldProcess = function(ctx) {
    var parsedReqUrl = urlUtil.parseUrl(ctx.req.url);
    return parsedReqUrl.pathname.indexOf(this.urlPathPrefix) === 0;
};

ServiceChannel.prototype.process = function(ctx) {
    var parsedReqUrl = urlUtil.parseUrl(ctx.req.url), resourcePathname = parsedReqUrl.pathname.replace(this.urlPathPrefix, "");
    if (resourcePathname === SERVICE_MSGS_PATHNAME) {
        this._processServiceMsg(ctx);
    } else {
        if (resourcePathname === TASK_SCRIPT_PATHNAME) {
            var referer = ctx.req.headers["referer"], refererInfo = referer && urlUtil.parseProxyUrl(referer);
            if (refererInfo) {
                var serviceChannel = this;
                refererInfo.originResourceInfo.url = urlUtil.formatUrl(refererInfo.originResourceInfo);
                this.cookieShelf.getClientCookieString(refererInfo.jobInfo, refererInfo.originResourceInfo.url, function(cookie) {
                    serviceChannel.events.for(refererInfo.jobInfo.ownerToken).emit("taskScriptRequested", refererInfo, cookie, function(taskScript) {
                        var script = serviceChannel.getSettingsScript({
                            cookie: cookie.replace(/'/g, "\\'"),
                            jobUid: refererInfo.jobInfo.uid,
                            jobOwnerToken: refererInfo.jobInfo.ownerToken,
                            serviceMsgUrl: serviceChannel.serviceMsgUrl,
                            originHost: refererInfo.originResourceInfo.host,
                            originProtocol: refererInfo.originResourceInfo.protocol,
                            originHostname: refererInfo.originResourceInfo.hostname,
                            originPort: refererInfo.originResourceInfo.port,
                            taskScript: taskScript
                        });
                        ctx.res.setHeader("content-type", "application/x-javascript");
                        ctx.res.setHeader("cache-control", "no-cache, no-store, must-revalidate");
                        ctx.res.setHeader("pragma", "no-cache");
                        ctx.res.end(script);
                    });
                });
            } else {
                ctx.res.statusCode = 404;
                ctx.res.end();
            }
        } else {
            var resourceDescr = this.staticResourceDescriptors[resourcePathname];
            if (resourceDescr) {
                ctx.res.setHeader("content-type", resourceDescr.contentType);
                ctx.res.setHeader("cache-control", "max-age=3600, public");
                ctx.res.end(resourceDescr.content);
            } else {
                ctx.res.statusCode = 404;
                ctx.res.end();
            }
        }
    }
};

ServiceChannel.prototype.registerStaticResource = function(resourcePathname, resourceDescr) {
    this.staticResourceDescriptors[resourcePathname] = resourceDescr;
    return this._getServiceUrl(resourcePathname);
};