var Ast = require("./ast"), Common = require("./common"), ErrCodes = require("./err_codes"), FuncCallValidator = require("./func_call_validator");

var StepsFacilityAnalyzer = module.exports = function(isMixin, rawStepData, errs) {
    this.rawStepData = rawStepData;
    this.isMixin = isMixin;
    this.errs = errs;
    this.ERR = {};
    this.ERR.MISPLACED_DECLARATION = isMixin ? ErrCodes.MISPLACED_MIXIN_DECLARATION : ErrCodes.MISPLACED_TEST_DECLARATION;
    this.ERR.EMPTY_NAME = isMixin ? ErrCodes.EMPTY_MIXIN_NAME : ErrCodes.EMPTY_TEST_NAME;
    this.ERR.DUPLICATE_NAME = isMixin ? ErrCodes.DUPLICATE_MIXIN_NAME : ErrCodes.DUPLICATE_TEST_NAME;
    this.ERR.INVALID_ASSIGNMENT = isMixin ? ErrCodes.INVALID_MIXIN_ASSIGNMENT : ErrCodes.INVALID_TEST_ASSIGNMENT;
    this.ERR.FACILITY_IS_EMPTY = isMixin ? ErrCodes.MIXIN_IS_EMPTY : ErrCodes.TEST_IS_EMPTY;
    this.ERR.MIXIN_USED_IN_MIXIN = ErrCodes.MIXIN_USED_IN_MIXIN;
    this.ERR.STEP_IS_NOT_A_FUNCTION = isMixin ? ErrCodes.MIXIN_STEP_IS_NOT_A_FUNCTION : ErrCodes.TEST_STEP_IS_NOT_A_FUNCTION_OR_MIXIN;
    this.ERR.STEP_IS_EMPTY = isMixin ? ErrCodes.MIXIN_STEP_IS_EMPTY : ErrCodes.TEST_STEP_IS_EMPTY;
};

StepsFacilityAnalyzer.prototype.run = function(astPath, filename) {
    if (!Ast.isPathMatch(Common.TEST_AND_MIXIN_DECLARATION_AST_PATH, astPath)) {
        this._err(this.ERR.MISPLACED_DECLARATION, filename, astPath);
        return;
    }
    var name = Ast.getAncestorByName("sub", astPath)[2][1];
    if (!name) {
        this._err(this.ERR.EMPTY_NAME, filename, astPath);
        return;
    }
    var isDuplicate = this.rawStepData[name];
    if (isDuplicate) {
        this._err(this.ERR.DUPLICATE_NAME, filename, astPath, {
            name: name
        });
        return;
    }
    var obj = Ast.getAncestorByName("assign", astPath)[3];
    if (Ast.getEntryName(obj) !== "object") {
        this._err(this.ERR.INVALID_ASSIGNMENT, filename, astPath);
        return;
    }
    var steps = obj[1];
    if (!steps.length) {
        this._err(this.ERR.FACILITY_IS_EMPTY, filename, astPath);
    } else {
        this._analyzeSteps(name, steps, filename);
    }
};

StepsFacilityAnalyzer.prototype._tryExtractTestCasesDirectiveAst = function(steps) {
    var firstStep = steps[0];
    if (firstStep[0] === Common.TEST_CASES_DIRECTIVE) {
        steps.splice(0, 1);
        return firstStep[1];
    }
    return null;
};

StepsFacilityAnalyzer.prototype._analyzeSteps = function(facilityName, steps, filename) {
    var analyzer = this;
    this.rawStepData[facilityName] = {
        names: [],
        asts: []
    };
    if (!this.isMixin) {
        this.rawStepData[facilityName].testCasesDirectiveAst = this._tryExtractTestCasesDirectiveAst(steps);
    }
    steps.forEach(function(step) {
        var stepAst = step[1], stepEntryName = Ast.getEntryName(stepAst);
        if (stepEntryName !== "function") {
            if (stepEntryName === "string") {
                var match = Common.DIRECTIVE_EXPRESSION_PATTERN.exec(stepAst[1]);
                if (match && match[1] === Common.MIXIN_INSERTION_POINT_DIRECTIVE_LVALUE) {
                    if (analyzer.isMixin) {
                        analyzer._err(analyzer.ERR.MIXIN_USED_IN_MIXIN, filename, stepAst);
                        return;
                    }
                    analyzer._insertStep(facilityName, step[0], {
                        isMixinInsertionPoint: true,
                        mixinName: match[2],
                        line: Ast.getCurrentSrcLineNum(stepAst)
                    });
                    return;
                }
            }
            analyzer._err(analyzer.ERR.STEP_IS_NOT_A_FUNCTION, filename, stepAst);
            return;
        }
        var funcBody = stepAst[3], funcLastEntr = funcBody[funcBody.length - 1];
        if (!funcLastEntr) {
            analyzer._err(analyzer.ERR.STEP_IS_EMPTY, filename, stepAst);
            return;
        }
        funcLastEntr[0].last = true;
        FuncCallValidator.validate(stepAst, filename, analyzer.errs, false);
        analyzer._insertStep(facilityName, step[0], stepAst);
    });
};

StepsFacilityAnalyzer.prototype._insertStep = function(facilityName, name, ast) {
    this.rawStepData[facilityName].names.push(name);
    this.rawStepData[facilityName].asts.push(ast);
};

StepsFacilityAnalyzer.prototype._err = function(code, filename, currentAst, additionalFields) {
    var line = Ast.getCurrentSrcLineNum(currentAst);
    this.errs.push(Common.createErrorObj(code, filename, line, additionalFields));
};