var util = require("util"), EventEmitter = require("events").EventEmitter, VirtualFS = require("../vfs"), errMsgBuilder = require("./err_msg_builder"), ERR = require("./server_errs");

var Api = exports.Api = function(testRunner, vfs, config, standalone) {
    EventEmitter.call(this);
    this.config = config;
    this.testRunner = testRunner;
    this.vfs = vfs;
    this.standalone = standalone;
    this._setupEvents();
};

util.inherits(Api, EventEmitter);

Api.DIRECTORY_TEST_TEXT = '"%s" directory tests';

Api.FIXTURE_TEST_TEXT = '"%s" fixture tests';

Api.SINGLE_TEST_TEXT = '"%s" single test';

Api.GROUP_TESTS_TEXT = '"%s" tests';

Api.DIR_TEST_SOURCE = "dir";

Api.FIXTURE_TEST_SOURCE = "fixture";

Api.SINGLE_TEST_SOURCE = "test";

Api.GROUP_TEST_SOURCE = "group";

Api.ROOT_MSG = "Root";

Api.prototype._getError = function(err) {
    return this.standalone ? err : errMsgBuilder.build(err);
};

Api.prototype._getTestSource = function(source, sourceType, suite) {
    var testUids = [], taskName = "";
    switch (sourceType) {
      case Api.SINGLE_TEST_SOURCE:
        if (suite.tests.hasOwnProperty(source)) {
            testUids = [ source ];
            taskName = util.format(Api.SINGLE_TEST_TEXT, suite.tests[source].name);
        } else {
            return {
                errs: this._getError({
                    code: ERR.API_TESTS_TARGET_NOT_EXIST,
                    source: source
                })
            };
        }
        break;

      case Api.FIXTURE_TEST_SOURCE:
        if (suite.fixtures.hasOwnProperty(source)) {
            testUids = suite.fixtures[source].tests;
            taskName = util.format(Api.FIXTURE_TEST_TEXT, suite.fixtures[source].name);
        } else {
            return {
                errs: this._getError({
                    code: ERR.API_TESTS_TARGET_NOT_EXIST,
                    source: source
                })
            };
        }
        break;

      case Api.DIR_TEST_SOURCE:
        var dirPath = source ? source.split("/") : [];
        testUids = VirtualFS.getTestUidsByPath(suite, dirPath);
        if (testUids.length) {
            taskName = util.format(Api.DIRECTORY_TEST_TEXT, dirPath[dirPath.length - 1] ? dirPath[dirPath.length - 1] : Api.ROOT_MSG);
        }
        break;

      case Api.GROUP_TEST_SOURCE:
        var group = [];
        for (var i = 0, length = source.length; i < length; i++) {
            var uid = source[i];
            if (suite.fixtures.hasOwnProperty(uid)) {
                group = group.concat(suite.fixtures[uid].tests);
            } else {
                if (suite.tests.hasOwnProperty(uid)) {
                    group.push(source[i]);
                } else {
                    return {
                        errs: this._getError({
                            code: ERR.API_TESTS_TARGET_NOT_EXIST,
                            source: uid
                        })
                    };
                }
            }
        }
        testUids = group;
        taskName = util.format(Api.GROUP_TESTS_TEXT, suite.tests[group[0]].group);
        break;

      default:
        return {
            errs: this._getError({
                code: ERR.API_UNKNOWN_FIXTURE_TYPE,
                type: sourceType
            })
        };
    }
    return {
        testUids: testUids,
        taskName: taskName
    };
};

Api.prototype._getActualWorkers = function(workers, browsers, taskName, callback) {
    var workerPool = this.testRunner.workerPool, workersList = [], errs = [];
    workers.forEach(function(workerName) {
        if (!workerPool.get(workerName)) {
            errs.push({
                code: ERR.API_TASK_WORKER_DOESNT_EXIST,
                taskName: taskName,
                workerName: workerName
            });
        } else {
            workersList.push(workerName);
        }
    });
    if (browsers) {
        workerPool.execBrowsersAndCreateWorkers(browsers, function(createErrs, createdWorkerNames) {
            if (createdWorkerNames) {
                workersList = workersList.concat(createdWorkerNames);
            }
            if (createErrs) {
                errs = errs.concat(createErrs);
            }
            callback(errs, workersList);
        });
    } else {
        callback(errs, workersList);
    }
};

Api.prototype._setupEvents = function() {
    var api = this;
    var taskUpdatedHandler = function(taskUid) {
        var report = api.testRunner.reporter.getTaskReportByUid(taskUid), task = api.testRunner.tasks[taskUid], status = report.status;
        api.testRunner.reporter.formatTaskReport(report, task.reportFormat, function(err, report) {
            if (status === "failed" || status === "succeeded") {
                if (!task.reportPath) {
                    api.emit("taskComplete", report);
                } else {
                    api.testRunner.reporter.saveReport(task.reportPath, report, function(err) {
                        if (err) {
                            throw errMsgBuilder.build(err);
                        }
                        api.emit("taskComplete", report);
                    });
                }
            } else {
                api.emit("taskUpdated", report);
            }
        });
    };
    this.testRunner.reporter.on("taskUpdated", function(task) {
        taskUpdatedHandler(task.id);
    });
    this.testRunner.reporter.on("taskDetailErrUpdated", function(task) {
        taskUpdatedHandler(task.id);
    });
    this.testRunner.workerPool.on("workerAdded", function(name) {
        api.emit("workerAdded", name);
    });
    this.testRunner.workerPool.on("workerDisconnected", function(workerName) {
        api.emit("workerDisconnected", workerName);
    });
};

Api.prototype.listConnectedWorkers = function() {
    return this.testRunner.workerPool.listConnectedWorkers();
};

Api.prototype.listAvailableBrowsers = function() {
    return Object.keys(this.config.browsers);
};

Api.prototype.listDirectory = function() {
    var path = null, vfsPath = [], callback = null;
    if (arguments.length === 2) {
        path = arguments[0];
        callback = arguments[1];
    } else {
        callback = arguments[0];
    }
    if (path && path !== "/") {
        vfsPath = path;
    }
    if (typeof vfsPath === "string") {
        vfsPath = VirtualFS.getPlatformDependentPath(vfsPath);
        vfsPath = vfsPath.split(path.sep);
    }
    this.vfs.getPathInfo(vfsPath, function(err, pathInfo) {
        callback(err, pathInfo);
    });
};

Api.prototype.runTests = function(options, callback) {
    var source = options.source || null, sourceType = options.sourceType || "dir", workers = options.workers || [], browsers = options.browsers || [], emulateCursor = options.emulateCursor || false, quarantineMode = options.quarantineMode || false, reportFormat = options.reportFormat || "json", reportPath = options.reportPath;
    var api = this, testRunner = this.testRunner, callbackExists = typeof callback === "function";
    if (!workers.length && !browsers.length) {
        if (callbackExists) {
            callback(ERR.API_NO_WORKERS_SPECIFIED_FOR_TASK, null, null);
        }
        return;
    }
    this.vfs.getSuite(function(suite) {
        var testSource = api._getTestSource(source, sourceType, suite);
        if (testSource.errs) {
            if (callbackExists) {
                callback(testSource.errs, null, null);
            }
        } else {
            var testUids = testSource.testUids, taskName = testSource.taskName, taskUid = null;
            if (testUids.length) {
                api._getActualWorkers(workers, browsers, taskName, function(errs, actualWorkers) {
                    if (!errs.length) {
                        errs = null;
                    }
                    if (actualWorkers.length) {
                        taskUid = testRunner.addTask(taskName, suite, testUids, actualWorkers, emulateCursor, quarantineMode, reportFormat, reportPath);
                        if (callbackExists) {
                            callback(errs, taskUid, actualWorkers);
                        }
                    } else {
                        if (callbackExists) {
                            callback(errs, null, null);
                        }
                    }
                });
            } else {
                if (callbackExists) {
                    callback([ api._getError({
                        code: ERR.API_NO_TESTS_TO_RUN
                    }) ], null, null);
                }
            }
        }
    });
};