var util = require("util"), errMsgBuilder = require("./../server/err_msg_builder"), moment = require("moment"), ejs = require("ejs"), path = require("path"), fs = require("fs"), ERR = require("./../server/server_errs"), EventEmitter = require("events").EventEmitter;

function hasTaskErrors(taskReport) {
    return Object.keys(taskReport.testErrReports).length > 0;
}

function getElapsedTime(startTime) {
    return startTime ? moment().diff(startTime, "seconds") : 0;
}

var Reporter = module.exports = function() {
    EventEmitter.call(this);
    this.taskReports = [];
};

util.inherits(Reporter, EventEmitter);

Reporter.JUNIT_XML_EJS_PATH = path.join(__dirname, "report_templates", "junit.xml.ejs");

Reporter.NUNIT_XML_EJS_PATH = path.join(__dirname, "report_templates", "nunit.xml.ejs");

Reporter.TIME_FORMAT = "MM/DD/YYYY h:mm:ss A";

Reporter.prototype.getTaskReportByUid = function(taskUid) {
    for (var i = 0; i < this.taskReports.length; i++) {
        if (this.taskReports[i].uid === taskUid) {
            this._updateTaskElapsedTime(this.taskReports[i]);
            return this.taskReports[i];
        }
    }
};

Reporter.prototype.getTaskReports = function() {
    var reporter = this;
    this.taskReports.forEach(function(taskReport) {
        reporter._updateTaskElapsedTime(taskReport);
    });
    return this.taskReports;
};

Reporter.prototype.taskAdded = function(taskUid, name, testCount, workerNames, reportUnstable) {
    var taskReport = {
        uid: taskUid,
        name: name,
        status: "pending",
        startedAt: null,
        completedAt: null,
        time: 0,
        workerNames: workerNames,
        testCount: testCount,
        failed: 0,
        passed: 0,
        testErrReports: {},
        passedTests: {}
    };
    if (reportUnstable) {
        taskReport.unstableTests = {};
    }
    this.completedRunsPerTest = {};
    this.taskReports.push(taskReport);
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.taskStarted = function(taskUid) {
    var taskReport = this.getTaskReportByUid(taskUid);
    taskReport.startedAt = moment().format(Reporter.TIME_FORMAT);
    if (taskReport.startedAt) {
        taskReport.startedAt = taskReport.startedAt.split(" ");
    }
    taskReport.status = "running";
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.taskCompleted = function(taskUid) {
    var taskReport = this.getTaskReportByUid(taskUid);
    taskReport.completedAt = taskReport.startedAt && moment().format(Reporter.TIME_FORMAT);
    if (taskReport.completedAt) {
        taskReport.completedAt = taskReport.completedAt.split(" ");
        taskReport.time = getElapsedTime(moment(taskReport.startedAt.join(" "), Reporter.TIME_FORMAT));
    }
    taskReport.status = hasTaskErrors(taskReport) ? "failed" : "succeeded";
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.testRunCompleted = function(testRun) {
    var taskReport = this.getTaskReportByUid(testRun.taskUid), testUid = testRun.test.uid;
    if (testRun.errs.length) {
        this._processTestRunErrs(testRun, taskReport, testUid);
        this.emit("taskDetailErrUpdated", {
            id: testRun.taskUid
        });
    }
    testRun.time = testRun.startTime ? getElapsedTime(moment(testRun.startTime, Reporter.TIME_FORMAT)) : 0;
    taskReport.time = taskReport.startedAt ? getElapsedTime(moment(taskReport.startedAt.join(" "), Reporter.TIME_FORMAT)) : 0;
    if (taskReport.unstableTests && testRun.quarantine.failed && testRun.quarantine.succeeded) {
        if (!taskReport.unstableTests[testUid]) {
            taskReport.unstableTests[testUid] = {
                name: testRun.test.name,
                fixtureName: testRun.fixture.name,
                fixturePath: testRun.fixture.path.join("/"),
                workerNames: []
            };
        }
        taskReport.unstableTests[testUid].workerNames.push(testRun.workerName);
    }
    if (taskReport.testErrReports[testUid]) {
        return;
    }
    if (!this.completedRunsPerTest[testRun.taskUid]) {
        this.completedRunsPerTest[testRun.taskUid] = {};
    }
    this.completedRunsPerTest[testRun.taskUid][testUid] = this.completedRunsPerTest[testRun.taskUid][testUid] || 0;
    if (++this.completedRunsPerTest[testRun.taskUid][testUid] === taskReport.workerNames.length) {
        taskReport.passed++;
        taskReport.passedTests[testRun.test.uid] = {
            fixtureName: testRun.fixture.name,
            fixturePath: testRun.fixture.path.join("/"),
            name: testRun.test.name,
            time: getElapsedTime(testRun.startTime)
        };
        if (taskReport.passed + taskReport.failed < taskReport.testCount) {
            this.emit("taskUpdated", {
                id: testRun.taskUid
            });
        }
    }
};

Reporter.prototype.formatTaskReport = function(report, reportFormat, callback) {
    switch (reportFormat) {
      case "junit":
        ejs.renderFile(Reporter.JUNIT_XML_EJS_PATH, {
            suite: report
        }, callback);
        break;

      case "nunit":
        ejs.renderFile(Reporter.NUNIT_XML_EJS_PATH, {
            suite: report
        }, callback);
        break;

      default:
        callback(null, report);
    }
};

Reporter.prototype.saveReport = function(reportPath, report, callback) {
    var savedReport = typeof report === "string" ? report : JSON.stringify(report, null, 4);
    fs.writeFile(reportPath, savedReport, function(err) {
        callback(err ? {
            code: ERR.REPORTER_WRITE_REPORT_FAILED,
            filePath: reportPath
        } : null);
    });
};

Reporter.prototype._processTestRunErrs = function(testRun, taskReport, testUid) {
    if (!taskReport.testErrReports[testUid]) {
        taskReport.testErrReports[testUid] = {
            name: testRun.test.name,
            fixtureName: testRun.fixture.name,
            fixturePath: testRun.fixture.path.join("/"),
            time: getElapsedTime(testRun.startTime),
            errs: []
        };
        taskReport.failed++;
        this.emit("taskListTestFailed", {
            id: testRun.taskUid
        });
    }
    testRun.errs.forEach(function(err) {
        var reportItem = {
            msg: errMsgBuilder.build(err.err),
            workerName: testRun.workerName,
            possibleCauses: []
        };
        err.possibleCauses.forEach(function(cause) {
            reportItem.possibleCauses.push({
                type: cause.type,
                msg: errMsgBuilder.build(cause.err)
            });
        });
        taskReport.testErrReports[testUid].errs.push(reportItem);
    });
};

Reporter.prototype._updateTaskElapsedTime = function(taskReport) {
    if (taskReport.status === "running" && taskReport.startedAt) {
        taskReport.time = getElapsedTime(moment(taskReport.startedAt.join(" "), Reporter.TIME_FORMAT));
    }
};